@@ -49,22 +49,31 @@
  */
 #include <stdio.h>
 #include <assert.h>
+#include <ctype.h>
 #include "tk.h"
 #include "dpInt.h"
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <netdb.h>
 
+#ifdef UNIX_SOCKET
 #include <sys/un.h>
+#endif
 #include <sys/param.h>
 #include <sys/ioctl.h>
+#ifndef NO_WRITEV
 #include <sys/uio.h>
+#endif
+#include <fcntl.h>
 
+extern int errno;
 #ifdef NO_ERRNO_H
-int errno;
-#define EINVAL 22
+#  define EAGAIN 11
+#  define EINVAL 22
+#  define EWOULDBLOCK 35
+#  define EPIPE 32
 #else
-#include <errno.h>
+#  include <errno.h>
 #endif
 
 #define max(a,b)	((a)>(b)?(a):(b))
@@ -82,9 +91,12 @@ int errno;
 static int  Tdp_inet_connect	_ANSI_ARGS_((char *host, int port,
 					     int server, 
 					     int udp));
+#ifdef UNIX_SOCKET
 static int  Tdp_unix_connect	_ANSI_ARGS_((char *path, 
 					     int server,
                                              int udp));
+#endif
+
 static void Tdp_HandleEvent	_ANSI_ARGS_((ClientData clientData, 
 					     int mask));
 
@@ -148,6 +160,7 @@ static unsigned char optFlags[MAX_OPEN_FILES];
 #define	FD_UNIX		8
 #define	FD_SERVER	16
 #define	FD_AUTO_CLOSE	32
+
 
 /*
  *--------------------------------------------------------------
@@ -171,16 +184,20 @@ Tdp_SetBlocking (fd, block)
     int fd;		/* File descriptor of socket */
     int block;		/* 1 if we should block from now on, 0 if not */
 {
-    int optval = !block;
+    int flags;
 
     if (block) {
 	if ((optFlags[fd] & FD_BLOCKING) == 0) {
-	    ioctl (fd, FIONBIO, &optval);
+	    fcntl (fd, F_GETFL, &flags);
+	    flags |= O_NDELAY;
+	    fcntl (fd, F_SETFL, &flags);
 	    optFlags[fd] |= FD_BLOCKING;
 	}
     } else {
 	if (optFlags[fd] & FD_BLOCKING) {
-	    ioctl (fd, FIONBIO, &optval);
+	    fcntl (fd, F_GETFL, &flags);
+	    flags &= ~O_NDELAY;
+	    fcntl (fd, F_SETFL, &flags);
 	    optFlags[fd] &= ~FD_BLOCKING;
 	}
     }
@@ -360,7 +377,7 @@ Tdp_FreeReadBuffer(fd)
 	readList = partial[fd];
 	partial[fd] = readList->next;
 	ckfree (readList->buffer);
-	ckfree (readList);
+	ckfree ((char *) readList);
     }
 }
 
@@ -389,6 +406,7 @@ Tdp_Unread (fd, buffer, numBytes, copy)
 {
     PartialRead *new;
 
+    if (numBytes == 0) return;
     new = (PartialRead *)ckalloc (sizeof(PartialRead));
     if (copy) {
 	new->buffer = ckalloc (numBytes);
@@ -466,7 +484,7 @@ Tdp_Read (fd, buffer, numReq, flags)
 	    tmp->offset += numToCopy;
 	    if (tmp->offset == tmp->bufSize) {
 		ckfree (tmp->buffer);
-		ckfree (tmp);
+		ckfree ((char *) tmp);
 		partial[fd] = readList;
 	    }
 	}
@@ -512,17 +530,23 @@ Tdp_ConnectCmd(notUsed, interp, argc, argv)
 {
     int fd;			/* Open file descriptor */
     int port;			/* User specified port number */
-    char *pathname, *host;	/* Pathname (unix) or host (inet) */
+    char *host;			/* Hostname (inet) */
+#ifdef UNIX_SOCKET
+    char *pathname;		/* Pathname (unix) */
     int unixSocket;		/* Unix domain socket? */
+#endif
     int udp;			/* UDP protocol? */
     int server;			/* Set up listening socket? */
     char tmp[256];
 
-    pathname = NULL;
     host = NULL;
     udp = 0;
     server = 0;
+
+#ifdef UNIX_SOCKET
     unixSocket = 0;
+    pathname = NULL;
+#endif
 
     if (argc < 2) {
 error:
@@ -535,10 +559,12 @@ error:
 			 (char *) NULL);
 	Tcl_AppendResult(interp, " \"dp_connect -udp port\"\n",
 			 (char *) NULL);
+#ifdef UNIX_SOCKET
 	Tcl_AppendResult(interp, " \"dp_connect -server path\"\n",
 			 (char *) NULL);
 	Tcl_AppendResult(interp, " or \"dp_connect path\"\n",
 			 (char *) NULL);
+#endif
 	return TCL_ERROR;
     }
 
@@ -558,7 +584,7 @@ error:
 	/*
 	 * Must be "dp_connect -udp port"
 	 */
-	if (argc != 3) {
+	if (argc < 3) {
 	    goto error;
 	}
 	host = "";
@@ -575,9 +601,13 @@ error:
 	}
 	host = "";
 	if (Tcl_GetInt(interp, argv[2], &port) != TCL_OK) {
+#ifdef UNIX_SOCKET
 	    pathname = argv[2];
 	    unixSocket = 1;
-	} 
+#else
+	    goto error;
+#endif
+        }
 
     } else {
 	/*
@@ -590,9 +620,11 @@ error:
 	    if (Tcl_GetInt(interp, argv[2], &port) != TCL_OK) {
 		return TCL_ERROR;
 	    }
+#ifdef UNIX_SOCKET
 	} else if (argc == 2) {
 	    pathname = argv[1];
 	    unixSocket = 1;
+#endif
 	} else {
 	    goto error;
 	}
@@ -602,14 +634,18 @@ error:
     /*
      * Create the connection
      */
+#ifdef UNIX_SOCKET
     if (unixSocket) {
 	fd = Tdp_unix_connect(pathname, server, udp);
-    } else {
+    } else
+#endif
+    {
 	fd = Tdp_inet_connect(host, port, server, udp);
     }
 
     if (fd < 0) {
 	/* Tell them why it fell apart */
+#ifdef UNIX_SOCKET
 	if (unixSocket) {
 	    if (server) {
 		Tcl_AppendResult (interp, 
@@ -621,7 +657,9 @@ error:
 			"Couldn't connect to \"", pathname, "\": ",
 			Tcl_PosixError(interp), (char *)NULL);
 	    }
-	} else if (server) {
+	} else
+#endif
+	if (server) {
 	    if (port == 0) {
 		Tcl_AppendResult (interp,
 			"Couldn't setup listening socket on any port: ",
@@ -649,7 +687,10 @@ error:
      */
     Tdp_FreeReadBuffer (fd);
 
-    if (!unixSocket) {
+#ifdef UNIX_SOCKET
+    if (!unixSocket)
+#endif
+	{
         struct sockaddr_in sockaddr;
         int res, len;
 	  
@@ -663,9 +704,12 @@ error:
         } else  {
 	    sprintf(tmp, "file%d %d", fd, ntohs(sockaddr.sin_port));
 	}
-    } else {
+    }
+#ifdef UNIX_SOCKET
+    else {
 	sprintf(tmp, "file%d", fd);
     }
+#endif
     Tcl_SetResult (interp, tmp, TCL_VOLATILE);
 
     Tdp_AllocateBuffer (fd);
@@ -674,9 +718,11 @@ error:
     } else {
 	optFlags[fd] = FD_TCP | FD_BLOCKING | FD_AUTO_CLOSE;
     }
+#ifdef UNIX_SOCKET
     if (unixSocket) {
 	optFlags[fd] |= FD_UNIX;
     }
+#endif
     if (server) {
 	optFlags[fd] |= FD_SERVER;
     }
@@ -730,7 +776,10 @@ Tdp_SocketOptionCmd (clientData, interp, argc, argv)
     }
     fd = fileno (filePtr);
 
-    c = tolower(argv[2][0]);
+    c = argv[2][0];
+    if (isupper(c))
+	c = tolower(c);
+
     len = strlen(argv[2]);
 
     /* ------------------------ SEND BUFFER ---------------------------- */
@@ -772,6 +821,7 @@ Tdp_SocketOptionCmd (clientData, interp, argc, argv)
     /* ------------------------ KEEP ALIVE ----------------------------- */
     } else if ((c == 'k') && (strncasecmp(argv[2], "keepAlive", len) == 0)) {
 	optname = SO_KEEPALIVE;
+	optval = 1;
 	if (argc == 4) {
 	    if (strcmp (argv[3], "yes") == 0) {
 		optval = 1;
@@ -802,6 +852,49 @@ Tdp_SocketOptionCmd (clientData, interp, argc, argv)
 	}
 	return TCL_OK;
 
+    /* ------------------------ LINGER --------------------------------- */
+    } else if ((c == 'l') && (strncasecmp(argv[2], "linger", len) == 0)) {
+	/* First int is on/off, second is linger time */
+	int linger[2];
+
+	if (argc == 3) {
+	    result = getsockopt(fd, SOL_SOCKET, SO_LINGER, (char *)linger,
+			        &optlen);
+	} else {
+
+	    if (Tcl_GetInt(interp, argv[3], &linger[1]) != TCL_OK) {
+		return TCL_ERROR;
+	    }
+	    if (linger[1] <= 0) {
+		linger[0] = linger[1] = 0;
+	    } else {
+		linger[0] = 1;
+	    }
+	    result = setsockopt(fd, SOL_SOCKET, SO_LINGER,
+				(char *)linger, sizeof(linger));
+	}
+	if (result == -1) {
+	    goto posixError;
+	}
+	sprintf (tmp, "%d", linger[0]?linger[1]:0);
+	Tcl_SetResult(interp, tmp, TCL_VOLATILE);
+	return TCL_OK;
+
+      /* ------------------------ RE-USE ADDR --------------------------- */
+      } else if ((c == 'r') && (strncasecmp(argv[2], "reuseAddr", len) == 0)) {
+  	optname = SO_REUSEADDR;
+  	if (argc == 4) {
+  	    if (strcmp (argv[3], "yes") == 0) {
+  		optval = 1;
+  	    } else if (strcmp (argv[3], "no") == 0) {
+  		optval = 0;
+  	    } else {
+  		Tcl_AppendResult (interp, argv[0], ": Bad value \"",
+  				  argv[3], "\"", NULL);
+  		return TCL_ERROR;
+  	    }
+  	}
+  
     /* ------------------------ ERROR ---------------------------------- */
     } else {
 	Tcl_AppendResult(interp, argv[0], "unknown option \"",
@@ -816,9 +909,7 @@ Tdp_SocketOptionCmd (clientData, interp, argc, argv)
     }
 
     if (result == -1) {
-	Tcl_AppendResult(interp, argv[0], ": ", Tcl_PosixError(interp),
-			 (char *) NULL);
-	return TCL_ERROR;
+	goto posixError;
     }
 
     if (optname == SO_KEEPALIVE) {
@@ -832,6 +923,11 @@ Tdp_SocketOptionCmd (clientData, interp, argc, argv)
 	Tdp_AllocateBuffer (fd);
     }
     return TCL_OK;
+
+posixError:
+    Tcl_AppendResult(interp, argv[0], ": ", Tcl_PosixError(interp),
+		     (char *) NULL);
+    return TCL_ERROR;
 }
 
 /*
@@ -1064,6 +1160,7 @@ Tdp_AcceptCmd(notUsed, interp, argc, argv)
     return TCL_OK;
 }
 
+#ifdef UNIX_SOCKET
 /*
  *----------------------------------------------------------------
  *
@@ -1116,6 +1213,7 @@ Tdp_unix_connect(path, server, udp)
     }
     return sock;
 }
+#endif
 
 /*
  *----------------------------------------------------------------
@@ -1152,7 +1250,11 @@ Tdp_inet_connect(host, port, server, udp)
 
     hostent = gethostbyname(host);
     if (hostent == NULL) {
-	hostaddr = inet_addr(host);
+	if (strlen(host) == 0) {
+	    hostaddr = 0;
+	} else {
+	    hostaddr = inet_addr(host);
+	}
 	if (hostaddr == -1) {
 	    if (server && !strlen(host)) {
 		hostaddr = INADDR_ANY;
@@ -1162,6 +1264,9 @@ Tdp_inet_connect(host, port, server, udp)
 	    }
 	}
 	_hostent.h_addr_list = (char **) hostaddrPtr;
+#ifdef CRAY_HACKS
+	hostaddr <<= 32;
+#endif
 	_hostent.h_addr_list[0] = (char *) &hostaddr;
 	_hostent.h_addr_list[1] = NULL;
 	_hostent.h_length = sizeof(hostaddr);
@@ -1179,11 +1284,23 @@ Tdp_inet_connect(host, port, server, udp)
     }
     memset((char *)&sockaddr, 0,sizeof(sockaddr));
     sockaddr.sin_family = AF_INET;
+#ifdef CRAY_HACKS
+{
+    unsigned long foo;
+
+    memcpy((char *)&foo,
+	   (char *)hostent->h_addr_list[0],
+	   (size_t)hostent->h_length);
+
+    sockaddr.sin_addr.s_addr = foo>>32;
+}
+#else
     memcpy((char *)&(sockaddr.sin_addr.s_addr),
 	   (char *)hostent->h_addr_list[0],
 	   (size_t)hostent->h_length);
-    sockaddr.sin_port = htons(port);
+#endif
 
+    sockaddr.sin_port = htons(port);
     if (server | udp) {
 	status = bind(sock, (struct sockaddr *)&sockaddr, sizeof(sockaddr));
     } else {
@@ -1747,6 +1864,7 @@ Tdp_SendCmd(clientData, interp, argc, argv)
 	newline = 0;
     }
 
+#ifndef NO_WRITEV
     {
       struct iovec	iov[2];
       register int	iovcnt = 1;
@@ -1760,6 +1878,22 @@ Tdp_SendCmd(clientData, interp, argc, argv)
       /* Use writev to reduce number of kernel calls */
       count = writev(fd, iov, iovcnt);
     }
+#else
+    /*
+     * Simulate writev() call with two write() calls, returning correct
+     * value.
+     */
+    {
+      char buf = '\n';
+      int len = strlen(argv[2]);
+
+      if (((count = write(fd, argv[2], len)) != -1) &&
+          ((count = write(fd, &buf, 1)) != -1)) {
+          count = len+1;
+      }
+    }
+#endif
+
     if (count == -1) {
 	if (errno == EPIPE) {
 	    /*
@@ -1812,6 +1946,7 @@ Tdp_PacketReceive(interp, fileHandle, peek)
     int packetLen;
     int fd;
     int headerSize;
+    unsigned char hbuf[8];
     int header[2];
     char *errMsg;
     int flags;
@@ -1846,10 +1981,10 @@ Tdp_PacketReceive(interp, fileHandle, peek)
     }
 
     /*
-     * Read in the header (2*sizeof(int))
+     * Read in the header (8 bytes)
      */
-    headerSize = 2*sizeof(int);
-    numRead = Tdp_Read (fd, (char *)header, headerSize, flags);
+    headerSize = 8;
+    numRead = Tdp_Read (fd, hbuf, headerSize, flags);
 
     if (numRead <= 0) {
 	goto readError;
@@ -1861,14 +1996,31 @@ Tdp_PacketReceive(interp, fileHandle, peek)
      */
     if (numRead < headerSize) {
 	if (!peek) {
-	    Tdp_Unread (fd, (char *)header, headerSize, 1);
+	    Tdp_Unread (fd, hbuf, numRead, 1);
 	}
 	Tcl_ResetResult(interp);
 	return TCL_OK;
     }
 
-    header[0] = ntohl(header[0]);
-    header[1] = ntohl(header[1]);
+    /*
+     * Convert header character stream into ints.  This works when the
+     * connecting machine has a different size int
+     * and takes care of endian problems.  It is also mostly
+     * backward compatible since network byte ordering (big endian) is
+     * used.
+     */
+
+    header[0] = 0;
+    header[0] |= hbuf[0] << 24;
+    header[0] |= hbuf[1] << 16;
+    header[0] |= hbuf[2] << 8;
+    header[0] |= hbuf[3];
+
+    header[1] = 0;
+    header[1] |= hbuf[4] << 24;
+    header[1] |= hbuf[5] << 16;
+    header[1] |= hbuf[6] << 8;
+    header[1] |= hbuf[7];
 
     /*
      * Format of each packet:
@@ -1880,7 +2032,7 @@ Tdp_PacketReceive(interp, fileHandle, peek)
     if (header[0] != PACKET_MAGIC) {
         Tcl_AppendResult(interp, "Error reading ", fileHandle,
 			 ": badly formatted packet", (char *) NULL);
-	return TCL_ERROR;
+	goto readError;
     }
     packetLen = header[1] - headerSize;
 
@@ -1920,15 +2072,25 @@ Tdp_PacketReceive(interp, fileHandle, peek)
     /*
      * We're not peeking, so we've consumed the header (this is normal mode).
      * Read in the packet, and if it's not all there, put it back.
+     *
+     * We have to be careful here, because we could block when if just
+     * the header came in (making the file readable at the beginning of this
+     * function) but the rest of the packet is still out on the network.
      */
-    numRead = Tdp_Read (fd, buffer, packetLen, flags);
-    if (numRead <= 0) {
+    if (Tdp_FDIsReady(fd) & TCL_FILE_READABLE) {
+	numRead = Tdp_Read (fd, buffer, packetLen, flags);
+    } else {
+	Tdp_Unread (fd, hbuf, headerSize, 1);
+	Tcl_ResetResult(interp);
+	return TCL_OK;
+    }
+    if (numRead < 0) {
 	goto readError;
     }
 
     if (numRead != packetLen) {
 	Tdp_Unread (fd, buffer, numRead, 1);
-	Tdp_Unread (fd, (char *)header, headerSize, 1);
+	Tdp_Unread (fd, hbuf, headerSize, 1);
 	Tcl_ResetResult(interp);
 	return TCL_OK;
     }
@@ -2010,7 +2172,7 @@ Tdp_PacketReceiveCmd(clientData, interp, argc, argv)
     if ((argc < 2) || (argc > 3)) {
       syntaxError:
 	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
-			 " fileId ?-peek?", (char *) NULL);
+			 "\" fileId ?-peek?", (char *) NULL);
 	return TCL_ERROR;
     }
 
@@ -2057,8 +2219,8 @@ Tdp_PacketSend(interp, fileHandle, message)
     int strLen;
     int packetLen;
     int numSent;
+    unsigned char hbuf[8];
     int header[2];
-    struct iovec iov[2];
     char tmp[256];
 
     if (Tcl_GetOpenFile(interp, fileHandle, 1, 1, &filePtr) != TCL_OK) {
@@ -2090,19 +2252,59 @@ Tdp_PacketSend(interp, fileHandle, message)
      *	  Next packetLen-(sizeof(int)) bytes are zero terminated message.
      */
     strLen = strlen (message);
-    packetLen = strLen + 2 * sizeof (int);
+    packetLen = strLen + 8;
 
-    header[0] = htonl (PACKET_MAGIC);
-    header[1] = htonl (packetLen);
+    header[0] = PACKET_MAGIC;
+    header[1] = packetLen;
+    /*
+     * Convert header ints to character stream.
+     * Network byte ordering (big endian) is used.
+     */
+
+    hbuf[0] = (header[0] & (unsigned long) 0xFF000000) >> 24;
+    hbuf[1] = (header[0] & (unsigned long) 0x00FF0000) >> 16;
+    hbuf[2] = (header[0] & (unsigned long) 0x0000FF00) >> 8;
+    hbuf[3] = (header[0] & (unsigned long) 0x000000FF);
+
+    hbuf[4] = (header[1] & (unsigned long) 0xFF000000) >> 24;
+    hbuf[5] = (header[1] & (unsigned long) 0x00FF0000) >> 16;
+    hbuf[6] = (header[1] & (unsigned long) 0x0000FF00) >> 8;
+    hbuf[7] = (header[1] & (unsigned long) 0x000000FF);
+
+#ifndef NO_WRITEV
+    {
+    struct iovec iov[2];
 
     /* Set up scatter/gather vector */
-    iov[0].iov_len = 2*sizeof(int);
-    iov[0].iov_base = (char *)header;
+    iov[0].iov_len = 8;
+    iov[0].iov_base = (char *)hbuf;
     iov[1].iov_len = strLen;
     iov[1].iov_base = message;
 
     /* Send it off, with error checking */
     numSent = writev (fd, iov, 2);
+    }
+#else
+    /*
+     * Again, simulate writev (this time using memcpy to put together
+     * the msg so it can go out in a single write() call
+     */
+    {
+    int len;
+    char *buffer;
+
+    len = strLen + 8;
+    buffer = (char *) ckalloc(len);
+
+    memcpy(buffer, hbuf, 8);
+    memcpy(buffer + 8, message, strLen);
+
+    numSent = write(fd, buffer, len);
+
+    ckfree(buffer);
+    }
+#endif
+
     if (numSent != packetLen) {
 
 	if ((errno == 0) || (errno == EWOULDBLOCK || errno == EAGAIN)) {
@@ -2110,7 +2312,7 @@ Tdp_PacketSend(interp, fileHandle, message)
 	     * Non-blocking I/O: return number of bytes actually sent.
 	     */
 	    Tcl_ResetResult(interp);
-	    sprintf (tmp, "%d", numSent - 2*sizeof(int));
+	    sprintf (tmp, "%d", numSent - 8);
 	    Tcl_SetResult(interp, tmp, TCL_VOLATILE);
 	    return TCL_OK;
 	} else if (errno == EPIPE) {
@@ -2135,7 +2337,7 @@ Tdp_PacketSend(interp, fileHandle, message)
     /*
      * Return the number of bytes sent (minus the header).
      */
-    sprintf (tmp, "%d", numSent - 2*sizeof(int));
+    sprintf (tmp, "%d", numSent - 8);
     Tcl_SetResult(interp, tmp, TCL_VOLATILE);
     return TCL_OK;
 }
@@ -2215,7 +2417,7 @@ Tdp_ReceiveFromCmd(notUsed, interp, argc, argv)
     struct sockaddr_in addr;
     int noaddr;
 
-    if ((argc < 2) || (argc > 4)) {
+    if ((argc < 2) || (argc > 5)) {
 	goto syntaxError;
     }
 
@@ -2232,7 +2434,7 @@ Tdp_ReceiveFromCmd(notUsed, interp, argc, argv)
 	c = argv[i][1];
 	if ((c == 'p') && (strncmp(argv[i], "-peek", len) == 0)) {
 	    flags |= MSG_PEEK;
-	} else if ((c == 'n') && (strncmp(argv[i], "-noaddr", len) != 0)) {
+	} else if ((c == 'n') && (strncmp(argv[i], "-noaddr", len) == 0)) {
 	    noaddr = 1;
 	} else if (Tcl_GetInt(interp, argv[i], &numBytes) != TCL_OK) {
 	    goto syntaxError;
@@ -2269,11 +2471,17 @@ Tdp_ReceiveFromCmd(notUsed, interp, argc, argv)
 	return TCL_ERROR;
     }
     if (!noaddr) {
-	addr.sin_port = ntohs(addr.sin_port);
-	addrName = Tdp_CreateAddress(addr.sin_addr.s_addr, addr.sin_port);
+	unsigned long the_addr;
+	int           the_port;
+
+	the_addr = addr.sin_addr.s_addr;
+	the_port = ntohs(addr.sin_port);
+
+	addrName = Tdp_CreateAddress(the_addr, the_port);
 	Tcl_SetResult(interp, addrName, TCL_STATIC);
 	Tcl_AppendElement(interp, buffer);
     } else {
+	buffer[count] = 0;
 	Tcl_SetResult(interp, buffer, TCL_STATIC);
     }
     return TCL_OK;
@@ -2281,7 +2489,7 @@ Tdp_ReceiveFromCmd(notUsed, interp, argc, argv)
   syntaxError:
     Tcl_AppendResult (interp,
                       "wrong # args: should be \"", argv[0],
-                      " fileId ?numBytes? ?-peek?\"",
+                      " fileId ?numBytes? ?-noaddr? ?-peek?\"",
                       (char *) NULL);
     return TCL_ERROR;
 }
