@@ -399,7 +399,7 @@ Tdp_CancelRPC (fd, errmsg)
     }
     if (errmsg) {
 	rpcValue[fd] = (char *)ckalloc(50);
-	sprintf (rpcValue[fd], "RPC Cancelled on file%d", fd);
+	sprintf (rpcValue[fd], "{RPC Cancelled on file%d} {}", fd);
     }
     
     if (rpcTimeoutReturn[fd]) {
@@ -507,8 +507,16 @@ Tdp_TimeoutHandler(clientData)
     if (rpcValue[fd]) {
 	ckfree((char *) rpcValue[fd]);
     }
-    rpcValue[fd] = (char *) ckalloc(strlen(interp->result) + 1);
-    strcpy(rpcValue[fd], interp->result);
+    if (rpcResult[fd] == TCL_ERROR) {
+	char *errorInfo;
+	errorInfo = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
+	rpcValue[fd] = (char *) ckalloc(strlen(errorInfo) +
+					strlen(interp->result) + 10);
+	sprintf (rpcValue[fd], "{%s} {%s}", interp->result, errorInfo);
+    } else {
+	rpcValue[fd] = (char *) ckalloc(strlen(interp->result) + 10);
+	strcpy (rpcValue[fd], interp->result);
+    }
 }
 
 /*
@@ -585,8 +593,17 @@ Tdp_ReceiveRPC_helper(interp, fileHandle, command, respond, fd)
     rd->checkCmd = checkCmd;
 
     if (respond) {
-	(void) send_tagged_packet(interp, fileHandle,
-		 (result != TCL_OK) ? TOK_ERR : TOK_RET, interp->result);
+	if (result != TCL_OK) {
+	    char *rv[2];
+	    char *str;
+	    rv[0] = interp->result;
+	    rv[1] = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
+	    str = Tcl_Merge(2, rv);
+	    send_tagged_packet(interp, fileHandle, TOK_ERR, str);
+	    ckfree (str);
+	} else {
+	    send_tagged_packet(interp, fileHandle, TOK_RET, interp->result);
+	}
     }
     Tcl_ResetResult(interp);
 
@@ -692,8 +709,8 @@ Tdp_ReceiveRPCCmd(clientData, interp, argc, argv)
  *
  *	TOK_RET	: message is the return value
  *		  of an RPC which this process requested.
- *	TOK_ERR	: message is the error return value
- *		  of an RPC which this process requested.
+ *	TOK_ERR	: message is a list off errorInfo and the error
+ *		  return value of an RPC which this process requested.
  *	TOK_RDO	: message is an incoming request for this
  *		  process to evaluate an RDO.
  *	TOK_RPC	: message is an incoming request for this
@@ -746,11 +763,14 @@ Tdp_ProcessRPCMessages(interp, fileHandle, wait)
 	     * Ignore any errors generated during cleanup, and return our own.
 	     */
 
+	    char *msg;
+	    msg = ckalloc (strlen(interp->result)+1);
+	    strcpy (msg, interp->result);
 	    Tdp_CleanupRPC (interp, fd, 0);
-
 	    Tcl_ResetResult (interp);
-	    Tcl_AppendResult (interp, "Error receiving RPC on ", fileHandle,
-			      " -- Closing connection", (char *)NULL);
+	    Tcl_SetResult (interp, msg, TCL_DYNAMIC);
+	    Tcl_AppendResult (interp, " caught while receiving RPC",
+			      "\n-- Closing connection", (char *)NULL);
 	    return result;
 	}
 
@@ -822,7 +842,7 @@ Tdp_ProcessRPCMessages(interp, fileHandle, wait)
 	    }
 
 	    rpcValue[fd] = (char *) ckalloc(strlen(buffer)+1);
-	    strcpy(rpcValue[fd], buffer);
+	    strcpy (rpcValue[fd], buffer);
 
 	    if (rpcTimeoutReturn[fd]) {
 		ckfree((char *) rpcTimeoutReturn[fd]);
@@ -1004,14 +1024,31 @@ Tdp_RPC(interp, fileHandle, command, events, timeout, timeoutReturn)
      */
 
     result = rpcResult[fd];
-
-    if (result == 0) {
+    if (result == TCL_ERROR) {
+	if (rpcValue[fd]) {
+	    char **rArgv;
+	    int rArgc;
+	    Tcl_SplitList(interp, rpcValue[fd], &rArgc, &rArgv);
+	    if (rArgc != 2) {
+		/* Probably old error protocol; do our best */
+		fprintf (stderr, "dp_RPC: detected pre-v3.1 error protocol\n");
+		Tcl_ResetResult (interp);
+		Tcl_SetResult (interp, rpcValue[fd], TCL_VOLATILE);
+	    } else {
+		Tcl_ResetResult (interp);
+		Tcl_AddErrorInfo (interp, rArgv[1]);
+		Tcl_SetResult(interp, rArgv[0], TCL_VOLATILE);
+	    }
+	    ckfree ((char *)rArgv);
+	    ckfree ((char *)rpcValue[fd]);
+	    rpcValue[fd] = NULL;
+	}
+    } else {
 	Tcl_ResetResult (interp);
-    }
-
-    if (rpcValue[fd]) {
-	Tcl_SetResult(interp, rpcValue[fd], TCL_DYNAMIC);
-	rpcValue[fd] = NULL;
+	if (rpcValue[fd]) {
+	    Tcl_SetResult(interp, rpcValue[fd], TCL_DYNAMIC);
+	    rpcValue[fd] = NULL;
+	}
     }
 
     /*
@@ -1255,6 +1292,7 @@ Tdp_ProcessRPCCommandCmd(unused, interp, argc, argv)
  */
 static char *ceCmdTemplate =
 "if [catch {%s} dp_rv] {\
+    dp_RDO $rpcFile set errorInfo \"$errorInfo\n    while remotely executing\n%s\"; \
     dp_RDO $rpcFile eval \"%s \\{$dp_rv\\}\"\
 } else {\
     dp_RDO $rpcFile eval \"%s $dp_rv\"\
@@ -1266,6 +1304,7 @@ static char *ceCmdTemplate =
  */
 static char *eCmdTemplate =
 "if [catch {%s} dp_rv] {\
+    dp_RDO $rpcFile set errorInfo \"$errorInfo\n    while remotely executing\n%s\"; \
     dp_RDO $rpcFile eval \"%s \\{$dp_rv\\}\"\
 }";
 
@@ -1291,6 +1330,7 @@ Tdp_RDOCmd(unused, interp, argc, argv)
     int result;
     char *callback;
     char *onerror;
+    char *cmdStr;
     int i;
 
     if (argc < 3) {
@@ -1305,14 +1345,18 @@ Tdp_RDOCmd(unused, interp, argc, argv)
     fileHandle = argv[1];
 
     callback = NULL;
-    onerror = NULL;
+    onerror = "tkerror";
     for (i=2; i<argc; i++) {
 	if (strcmp(argv[i], "-callback") == 0) {
 	    i++;
 	    callback = argv[i];
 	} else if (strcmp(argv[i], "-onerror") == 0) {
 	    i++;
-	    onerror = argv[i];
+	    if (strcmp(argv[i], "none") == 0) {
+		onerror = NULL;
+	    } else {
+		onerror = argv[i];
+	    }
 	} else {
 	    break;
 	}
@@ -1334,11 +1378,14 @@ Tdp_RDOCmd(unused, interp, argc, argv)
 	     *	   dp_RDO $rpcFile eval {$callback $dp_rv}
 	     *  }
 	     */
+	    cmdStr = Tcl_Merge(argc, argv);
 	    command = (char *) ckalloc(strlen(cmd) +
+				       strlen(cmdStr) +
 				       strlen(onerror) +
 				       strlen(callback) +
 				       strlen(ceCmdTemplate));
-	    sprintf (command, ceCmdTemplate, cmd, onerror, callback);
+	    sprintf (command, ceCmdTemplate, cmd, cmdStr, onerror, callback);
+	    ckfree (cmdStr);
 	} else {
 	    /*
 	     * Just onerror specified.  Form of
@@ -1347,10 +1394,13 @@ Tdp_RDOCmd(unused, interp, argc, argv)
 	     *	   dp_RDO $rpcFile eval {$onerror $dp_rv}
 	     *  }
 	     */
+	    cmdStr = Tcl_Merge(argc, argv);
 	    command = (char *) ckalloc(strlen(cmd) +
 				       strlen(onerror) +
+				       strlen(cmdStr) +
 				       strlen(eCmdTemplate));
-	    sprintf (command, eCmdTemplate, cmd, onerror);
+	    sprintf (command, eCmdTemplate, cmd, cmdStr, onerror);
+	    ckfree (cmdStr);
 	}
     } else {
 	if (callback != NULL) {
@@ -1537,7 +1587,9 @@ Tdp_RPCInit(interp)
     Tcl_CallWhenDeleted(interp,
 			(Tcl_InterpDeleteProc *)Tdp_FreeRPCInterpData,
 			(ClientData)rd);
-    Tcl_CreateTrace(interp, INT_MAX, Tdp_CommandTrace, (ClientData)rd);
+#define BYTES_PER_INT	(sizeof(int)/sizeof(char))
+#define LARGE_INT	(1 << (8*BYTES_PER_INT-2))
+    Tcl_CreateTrace(interp, LARGE_INT, Tdp_CommandTrace, (ClientData)rd);
 
     /*
      * Create commands.
